@@ -210,6 +210,7 @@ long amnt, nlines;
  b->changed=0;
  b->count=1;
  b->name=0;
+ b->er= -3;
  b->bof=palloc(); izque(P,link,b->bof);
  b->bof->end=0;
  b->bof->b=b;
@@ -384,19 +385,18 @@ P *p;
  if(piseof(p)) return 1;
  c=brc(p);
  if(c=='\n') return 1;
-#ifdef __MSDOS__
- if(c=='\r')
-  {
-  P *q=pdup(p);
-  pfwrd(q,1L);
-  if(pgetc(q)=='\n')
+ if(p->b->o.crlf)
+  if(c=='\r')
    {
-   prm(q);
-   return 1;
+   P *q=pdup(p);
+   pfwrd(q,1L);
+   if(pgetc(q)=='\n')
+    {
+    prm(q);
+    return 1;
+    }
+   else prm(q);
    }
-  else prm(q);
-  }
-#endif
  return 0;
  }
 
@@ -498,13 +498,11 @@ P *p;
  if(++p->ofst==GSIZE(p->hdr)) pnext(p); 
  ++p->byte;
  if(c=='\n') ++p->line, p->col=0, p->valcol=1;
-#ifdef __MSDOS__
- else if(c=='\r')
+ else if(p->b->o.crlf && c=='\r')
   {
   if(brc(p)=='\n') return pgetc(p);
   else ++p->col;
   }
-#endif
  else
   {
   if(c=='\t') p->col+=p->b->o.tab-p->col%p->b->o.tab;
@@ -561,15 +559,13 @@ int prgetc(p)
 P *p;
  {
  int c=prgetc1(p);
-#ifdef __MSDOS__
- if(c=='\n')
+ if(p->b->o.crlf && c=='\n')
   {
   c=prgetc1(p);
   if(c=='\r') return '\n';
   if(c!=MAXINT) pgetc(p);
   c='\n';
   }
-#endif
  return c;
  }
 
@@ -628,25 +624,24 @@ P *p;
 P *peol(p)
 P *p;
  {
-#ifdef __MSDOS__
- while(!piseol(p)) pgetc(p);
-#else
- while(p->ofst!=GSIZE(p->hdr))
-  {
-  unsigned char c;
-  if(p->ofst>=p->hdr->hole) c=p->ptr[p->ofst+p->hdr->ehole-p->hdr->hole];
-  else c=p->ptr[p->ofst];
-  if(c=='\n') break;
-  else
+ if(p->b->o.crlf)
+  while(!piseol(p)) pgetc(p);
+ else
+  while(p->ofst!=GSIZE(p->hdr))
    {
-   ++p->byte;
-   ++p->ofst;
-   if(c=='\t') p->col+=p->b->o.tab-p->col%p->b->o.tab;
-   else ++p->col;
-   if(p->ofst==GSIZE(p->hdr)) pnext(p); 
+   unsigned char c;
+   if(p->ofst>=p->hdr->hole) c=p->ptr[p->ofst+p->hdr->ehole-p->hdr->hole];
+   else c=p->ptr[p->ofst];
+   if(c=='\n') break;
+   else
+    {
+    ++p->byte;
+    ++p->ofst;
+    if(c=='\t') p->col+=p->b->o.tab-p->col%p->b->o.tab;
+    else ++p->col;
+    if(p->ofst==GSIZE(p->hdr)) pnext(p); 
+    }
    }
-  }
-#endif
  return p;
  }
 
@@ -694,13 +689,11 @@ P *p;
   }
   while(c!='\n');
  --p->line;
-#ifdef __MSDOS__
- if(c=='\n')
+ if(p->b->o.crlf && c=='\n')
   {
   int k=prgetc1(p);
   if(k!='\r' && k!=MAXINT) pgetc(p);
   }
-#endif
  return p;
  }
 
@@ -734,9 +727,7 @@ long goalcol;
   if(p->ofst>=p->hdr->hole) c=p->ptr[p->ofst+p->hdr->ehole-p->hdr->hole];
   else c=p->ptr[p->ofst];
   if(c=='\n') break;
-#ifdef __MSDOS__
-  if(c=='\r' && piseol(p)) break;
-#endif
+  if(p->b->o.crlf && c=='\r' && piseol(p)) break;
   if(c=='\t') wid=p->b->o.tab-p->col%p->b->o.tab;
   else wid=1;
   if(p->col+wid>goalcol) break;
@@ -1465,12 +1456,12 @@ int hdramnt;
 
  for(pp=p->link.next;pp!=p;pp=pp->link.next)
   if(pp->line==p->line &&
-     (pp->byte>p->byte || pp->end)) pp->valcol=0;
+     (pp->byte>p->byte || pp->end && pp->byte==p->byte)) pp->valcol=0;
  for(pp=p->link.next;pp!=p;pp=pp->link.next)
   if(pp->byte==p->byte && !pp->end)
    if(pp->ptr) pset(pp,p);
    else poffline(pset(pp,p));
-  else if(pp->byte>p->byte || pp->end)
+  else if(pp->byte>p->byte || pp->end && pp->byte==p->byte)
    {
    pp->byte+=amnt;
    pp->line+=nlines;
@@ -1539,11 +1530,8 @@ P *binsc(p,c)
 P *p;
 char c;
  {
-#ifdef __MSDOS__
- if(c=='\n') return binsm(p,"\r\n",2);
- else
-#endif
- return binsm(p,&c,1);
+ if(p->b->o.crlf && c=='\n') return binsm(p,"\r\n",2);
+ else return binsm(p,&c,1);
  }
 
 P *binss(p,s)
@@ -1683,6 +1671,7 @@ char *s;
  B *b;
  long skip,amnt;
  char *n;
+ int nowrite=0;
 
  if(!s || !s[0])
   {
@@ -1690,6 +1679,7 @@ char *s;
   b=bmk(NULL);
   setopt(&b->o,"");
   b->rdonly=b->o.readonly;
+  b->er=error;
   return b;
   }
 
@@ -1707,7 +1697,14 @@ char *s;
  else
 #endif
  if(!zcmp(n,"-")) fi=stdin;
- else fi=fopen(n,"r");
+ else
+  {
+  fi=fopen(n,"r+");
+  if(!fi) nowrite=1;
+  else fclose(fi);
+  fi=fopen(n,"r");
+  if(!fi) nowrite=0;
+  }
  joesep(n);
 
  /* Abort if couldn't open */
@@ -1758,10 +1755,12 @@ char *s;
  if(error || s[0]=='!' || skip || amnt!=MAXLONG) b->backup=1, b->changed=0;
  else if(!zcmp(n,"-")) b->backup=1, b->changed=1;
  else b->backup=0, b->changed=0;
+ if(nowrite) b->rdonly=b->o.readonly=1;
 
  /* Eliminate parsed name */
  vsrm(n);
 
+ b->er=error;
  return b;
  }
 
@@ -1778,6 +1777,7 @@ char *s;
   setopt(&b->o,"");
   b->rdonly=b->o.readonly;
   b->internal=0;
+  b->er=error;
   return b;
   }
  for(b=bufs.link.next;b!=&bufs;b=b->link.next)
