@@ -21,20 +21,15 @@ JOE; see the file COPYING.  If not, write to the Free Software Foundation,
 /* These should exist on every UNIX system */
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <sys/ioctl.h>
 #include <stdio.h>
 #include <signal.h>
 #include <fcntl.h>
 #include <errno.h>
 extern int errno;
 
-/* These depend on which tty structure the UNIX uses */
+int idleout=1;
 
-#ifndef TTYPOSIX
-#ifndef TTYSV
 #include <sys/param.h>
-#endif
-#endif
 
 #include "config.h"
 
@@ -43,15 +38,24 @@ extern int errno;
  * we actually have.
  */
 #ifdef TTYPOSIX
+
+#ifdef SYSPOSIX
+#include <sys/termios.h>
+#else
 #include <termios.h>
-char *getcwd(); char *pwd() { static char buf[1024]; return getcwd(buf,1024); }
+#endif
+
 #else
 #ifdef TTYSV
+
+#ifdef SYSSV
+#include <sys/termio.h>
+#else
 #include <termio.h>
-char *getcwd(); char *pwd() { static char buf[1024]; return getcwd(buf,1024); }
+#endif
+
 #else
 #include <sgtty.h>
-char *getwd(); char *pwd() { static char buf[1024]; return getwd(buf); }
 #endif
 #endif
 
@@ -69,11 +73,15 @@ char *getwd(); char *pwd() { static char buf[1024]; return getwd(buf); }
 
 /* I'm not sure if SCO_UNIX and ISC have __svr4__ defined, but I think
    they might */
-#ifdef SCO_UNIX
+#ifdef M_SYS5
+#ifndef M_XENIX
+#include <sys/stream.h>
+#include <sys/ptem.h>
 #ifndef __svr4__
 #define __svr4__ 1
 #endif
 #endif
+#endif
 
 #ifdef ISC
 #ifndef __svr4__
@@ -81,17 +89,7 @@ char *getwd(); char *pwd() { static char buf[1024]; return getwd(buf); }
 #endif
 #endif
 
-#ifdef sgi
-#ifndef __svr4__
-#define __svr4__ 1
-#endif
-#endif
-
 #ifdef __svr4__
-/* I don't think these two are needed if you have 'stropts' (sgi doesn't
- * even have them). */
-/* #include <sys/stream.h> */
-/* #include <sys/ptem.h> */
 #include <stropts.h>
 #endif
 
@@ -101,6 +99,17 @@ char *getwd(); char *pwd() { static char buf[1024]; return getwd(buf); }
 #include "path.h"
 #include "tty.h"
 
+/* The pwd function */
+#ifdef TTYPOSIX
+char *getcwd(); char *pwd() { static char buf[1024]; return getcwd(buf,1024); }
+#else
+#ifdef TTYSV
+char *getcwd(); char *pwd() { static char buf[1024]; return getcwd(buf,1024); }
+#else
+char *getwd(); char *pwd() { static char buf[1024]; return getwd(buf); }
+#endif
+#endif
+
 /** Aliased defines **/
 
 /* O_NDELAY, O_NONBLOCK, and FNDELAY are all synonyms for placing a descriptor
@@ -341,11 +350,9 @@ void ttopnn()
 #endif
  
  if(!termin)
-#ifdef IDLEOUT
-  if(!(termin=stdin) || !(termout=stdout))
-#else
-  if(!(termin=fopen("/dev/tty","r")) || !(termout=fopen("/dev/tty","w")))
-#endif
+  if(idleout ? (!(termin=stdin) || !(termout=stdout)) :
+               (!(termin=fopen("/dev/tty","r")) || 
+               !(termout=fopen("/dev/tty","w"))))
    {
    fprintf(stderr,"Couldn\'t open /dev/tty\n");
    exit(1);
@@ -366,19 +373,19 @@ void ttopnn()
  tcgetattr(fileno(termin),&oldterm);
  newterm=oldterm;
  newterm.c_lflag=0;
- if(noxon)  newterm.c_iflag&=~(ICRNL|IGNCR|INLCR|IXON);
+ if(noxon)  newterm.c_iflag&=~(ICRNL|IGNCR|INLCR|IXON|IXOFF);
  else newterm.c_iflag&=~(ICRNL|IGNCR|INLCR);
  newterm.c_oflag=0;
  newterm.c_cc[VMIN]=1;
  newterm.c_cc[VTIME]=0;
- tcsetattr(fileno(termin),TCSANOW,&newterm);
+ tcsetattr(fileno(termin),TCSADRAIN,&newterm);
  bbaud=cfgetospeed(&newterm);
 #else
 #ifdef TTYSV
  ioctl(fileno(termin),TCGETA,&oldterm);
  newterm=oldterm;
  newterm.c_lflag=0;
- if(noxon)  newterm.c_iflag&=~(ICRNL|IGNCR|INLCR|IXON);
+ if(noxon)  newterm.c_iflag&=~(ICRNL|IGNCR|INLCR|IXON|IXOFF);
  else newterm.c_iflag&=~(ICRNL|IGNCR|INLCR);
  newterm.c_oflag=0;
  newterm.c_cc[VMIN]=1;
@@ -443,7 +450,7 @@ void ttclsn()
  ttflsh();
  
 #ifdef TTYPOSIX
- tcsetattr(fileno(termin),TCSANOW,&oldterm);
+ tcsetattr(fileno(termin),TCSADRAIN,&oldterm);
 #else
 #ifdef TTYSV
  ioctl(fileno(termin),TCSETAW,&oldterm);
@@ -747,6 +754,19 @@ void mpxend()
  * string in static buffer or NULL if couldn't get a pair.
  */
 
+#ifdef sgi
+
+/* Newer sgi machines can do it the __svr4__ way, but old ones can't */
+
+extern char *_getpty();
+
+char *getpty(ptyfd)
+int *ptyfd;
+ {
+ return _getpty(ptyfd,O_RDWR,0600,0);
+ }
+
+#else
 #ifdef __svr4__
 
 /* Strange streams way */
@@ -765,6 +785,7 @@ int *ptyfd;
  }
 
 #else
+
 /* The normal way: for each possible pty/tty pair, try to open the pty and
  * then the corresponding tty.  If both could be opened, close them both and
  * then re-open the pty.  If that succeeded, return with the opened pty and the
@@ -826,6 +847,7 @@ int *ptyfd;
  }
 
 #endif
+#endif
 
 int dead=0;
 
@@ -847,6 +869,29 @@ struct sigvec inew={death,0,SV_INTERRUPT};
 #endif
 #endif
 
+/* Build a new environment */
+
+extern char **mainenv;
+
+char **newenv(old,s)
+char **old, *s;
+ {
+ char **new;
+ int x,y,z;
+ for(x=0;old[x];++x);
+ new=(char **)malloc((x+2)*sizeof(char *));
+ for(x=0,y=0;old[x];++x)
+  {
+  for(z=0;s[z]!='=';++z) if(s[z]!=old[x][z]) break;
+  if(s[z]=='=')
+   { if(s[z+1]) new[y++]=s; }
+  else new[y++]=old[x];
+  }
+ if(x==y) new[y++]=s;
+ new[y]=0;
+ return new;
+ }
+
 MPX *mpxmk(ptyfd,cmd,args,func,object,die,dieobj)
 int *ptyfd;
 char *cmd;
@@ -909,6 +954,7 @@ void *dieobj;
   
    if((x=open(name,O_RDWR))!= -1)    /* Standard input */
     {
+    char **env=newenv(mainenv,"TERM=");
 #ifdef __svr4__
     ioctl(x,I_PUSH,"ptem");
     ioctl(x,I_PUSH,"ldterm");
@@ -920,7 +966,7 @@ void *dieobj;
     /* We could probably have a special TTY set-up for JOE, but for now
      * we'll just use the TTY setup for the TTY was was run on */
 #ifdef TTYPOSIX
-    tcsetattr(0,TCSANOW,&oldterm);
+    tcsetattr(0,TCSADRAIN,&oldterm);
 #else
 #ifdef TTYSV
     ioctl(0,TCSETAW,&oldterm);
@@ -932,7 +978,7 @@ void *dieobj;
 #endif
   
     /* Execute the shell */
-    execv(cmd,args);
+    execve(cmd,args,env);
     }
 
    _exit(0);
